import math
from functools import partial
import torch
import torch.nn.functional as F
import numpy as np
from torch.autograd import Variable
from baseline.utils import Offsets, exporter
from eight_mile.pytorch.layers import repeat_batch, gnmt_length_penalty, BeamSearchBase, rnn_cell, WeightTieDense, subsequent_mask, TransformerDecoderStack
from baseline.model import register_arc_policy, register_decoder, create_seq2seq_arc_policy
from baseline.pytorch.seq2seq.encoders import TransformerEncoderOutput
from baseline.pytorch.torchy import (
    tie_weight,
    pytorch_linear,
    LuongDotProductAttention,
    BahdanauAttention,
    ScaledDotProductAttention,
    LuongGeneralAttention,
)

__all__ = []
export = exporter(__all__)


class ArcPolicy(torch.nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, encoder_outputs, hsz, beam_width=1):
        pass


class AbstractArcPolicy(ArcPolicy):

    def __init__(self):
        super().__init__()

    def get_state(self, encoder_outputs):
        pass

    def forward(self, encoder_output, hsz, beam_width=1):
        h_i = self.get_state(encoder_output)
        context = encoder_output.output
        if beam_width > 1:
            with torch.no_grad():
                context = repeat_batch(context, beam_width)
                if type(h_i) is tuple:
                    h_i = repeat_batch(h_i[0], beam_width, dim=1), repeat_batch(h_i[1], beam_width, dim=1)
                else:
                    h_i = repeat_batch(h_i, beam_width, dim=1)
        batch_size = context.shape[0]
        h_size = (batch_size, hsz)
        with torch.no_grad():
            init_zeros = context.data.new(*h_size).zero_()
        return h_i, init_zeros, context


@register_arc_policy(name='default')
class TransferLastHiddenPolicy(AbstractArcPolicy):

    def __init__(self):
        super().__init__()

    def get_state(self, encoder_outputs):
        return encoder_outputs.hidden


@register_arc_policy(name='no_arc')
class NoArcPolicy(AbstractArcPolicy):

    def __init__(self):
        super().__init__()

    def get_state(self, encoder_outputs):
        final_encoder_state = encoder_outputs.hidden
        if isinstance(final_encoder_state, tuple):
            s1, s2 = final_encoder_state
            return s1 * 0, s2 * 0
        return final_encoder_state * 0


@register_decoder(name='vanilla')
class RNNDecoder(torch.nn.Module):

    def __init__(self, tgt_embeddings, **kwargs):
        """Construct an RNN decoder.  It provides the input size, the rest is up to the impl.

        The default implementation provides an RNN cell, followed by a linear projection, out to a softmax

        :param input_dim: The input size
        :param kwargs:
        :return: void
        """
        super().__init__()
        self.hsz = kwargs['hsz']
        self.arc_policy = create_seq2seq_arc_policy(**kwargs)
        self.tgt_embeddings = tgt_embeddings
        rnntype = kwargs.get('rnntype', 'lstm')
        layers = kwargs.get('layers', 1)
        feed_input = kwargs.get('feed_input', True)
        dsz = tgt_embeddings.get_dsz()
        if feed_input:
            self.input_i = self._feed_input
            dsz += self.hsz
        else:
            self.input_i = self._basic_input
        pdrop = kwargs.get('dropout', 0.5)
        self.decoder_rnn = rnn_cell(dsz, self.hsz, rnntype, layers, pdrop)
        self.dropout = torch.nn.Dropout(pdrop)
        self.init_attn(**kwargs)

        do_weight_tying = bool(kwargs.get('tie_weights', True))

        if do_weight_tying:
            if self.hsz != self.tgt_embeddings.get_dsz():
                raise ValueError("weight tying requires hsz == embedding dsz, got {} hsz and {} dsz".format(self.hsz, self.tgt_embeddings.get_dsz()))
            self.preds = WeightTieDense(self.tgt_embeddings)
        else:
            self.preds = pytorch_linear(self.hsz, self.tgt_embeddings.get_vsz())

    @staticmethod
    def _basic_input(dst_embed_i, _):
        """
        In this function the destination embedding is passed directly to into the decoder.  The output of previous H
        is ignored.  This is implemented using a bound method to a field in the class for speed so that this decision
        is handled at initialization, not as a conditional in the training or inference

        :param embed_i: The embedding at i
        :param _: Ignored
        :return: basic input
        """
        return dst_embed_i.squeeze(0)

    @staticmethod
    def _feed_input(embed_i, attn_output_i):
        """
        In this function the destination embedding is concatenated with the previous attentional output and
        passed to the decoder. This is implemented using a bound method to a field in the class for speed
        so that this decision is handled at initialization, not as a conditional in the training or inference

        :param embed_i: The embedding at i
        :param output_i: This is the last H state
        :return: an input that is a concatenation of previous state and destination embedding
        """
        embed_i = embed_i.squeeze(0)
        return torch.cat([embed_i, attn_output_i], 1)

    def forward(self, encoder_outputs, dst):
        src_mask = encoder_outputs.src_mask
        h_i, output_i, context_bth = self.arc_policy(encoder_outputs, self.hsz)
        output_tbh, _ = self.decode_rnn(context_bth, h_i, output_i, dst.transpose(0, 1), src_mask)
        pred = self.output(output_tbh)
        return pred.transpose(0, 1).contiguous()

    def decode_rnn(self, context_bth, h_i, output_i, dst_tb, src_mask):
        """Decode some steps of the RNN

        This function is used to take steps on an RNN. It produces an output feature for
        each time step. The number of time steps is controlled by the size of the dst_tbh
        parameter. It can be used durning training to get the output for an entire sentence
        via teacher forcing by passing in the lagged targets as dst_tbh or it can be used for
        decoding by making multiple calls, where each dst_tbh is a single timestep created
        by selecting a token from the last time step.

        In baseline we call the tokens that we want to produce at time step `t` the `tgt`.
        The token that was produced at time step `t-1` is called the `dst`. The values of
        `dst` can either come from a pre-defined tensor in the case of teacher forcing
        or a from the argmax of the previous output in the case of decoding.

        Note:
            We have switched the majority of the seq2seq components to be batch first
            but this is still a hold out in that it is mostly time first. There is a
            weird mismatch right now because the context is batch first but the dst
            inputs and the output values are time first. I didn't want to change the whole
            seq2seq sections that relay on time first becuase it seemed like it would
            cause breakage and instead decided to explicity document it here.

        :param context_bth: The encoder outputs [B, T, H]
        :param h_i: The current hidden state of the RNN. Tuple[[L, B, H], [L, B, H]] where
            L is the number of layers
        :param output_i: The output features from the previous time step, [B, H]
        :param dst_tb: The input tokens from the previous time steps. [T, B]
        :param src_mask: The mask used for calculate valid attention score when looking at
            encoder outputs. [B, T]

        :Returns:
            Tuple[Torch.Tensor[T, B, H], Tuple[[L, B, H], [L, B, H]]
        """
        # Embed the `dst` values. These are often in the shape of [T, B] during training
        # where the whole sequence is known up front and the shape of [1, B] during
        # inference where we are decoding a single step at a time.
        embed_out_tbh = self.tgt_embeddings(dst_tb)

        outputs = []

        # Iterate through the `dst` embeddings one at a time. The reason for doing this at inference
        # time is obvious but we do it at training time too so that we run attention.
        for i, embed_i in enumerate(embed_out_tbh.split(1)):
            # Input feeding would use previous attentional output in addition to destination embeddings
            embed_i = self.input_i(embed_i, output_i)
            # Run the RNN for a single step on the current input and the previous hidden state.
            # Save this hidden state for use next time.
            output_i, h_i = self.decoder_rnn(embed_i, h_i)
            # Run attention between the RNN output (at this time step) and the encoder output.
            output_i = self.attn(output_i, context_bth, src_mask)
            output_i = self.dropout(output_i)
            # Save our outputs into a list that will be turned into a tensor later
            outputs.append(output_i)

        # Stack all the outputs into a single [T, B, H] tensor
        outputs_tbh = torch.stack(outputs)
        # Return the hidden state too for use next time.
        return outputs_tbh, h_i

    def attn(self, output_t, context, src_mask=None):
        return output_t

    def init_attn(self, **kwargs):
        pass

    def output(self, x):
        pred = F.log_softmax(self.preds(x.view(x.size(0)*x.size(1), -1)), dim=-1)
        pred = pred.view(x.size(0), x.size(1), -1)
        return pred

    class BeamSearch(BeamSearchBase):

        def __init__(self, parent, **kwargs):
            super().__init__(**kwargs)
            self.parent = parent

        def init(self, encoder_outputs):
            """Tile batches for encoder inputs and the likes."""
            src_mask = repeat_batch(encoder_outputs.src_mask, self.K)
            h_i, dec_out, context = self.parent.arc_policy(encoder_outputs, self.parent.hsz, self.K)
            return h_i, dec_out, context, src_mask

        def step(self, paths, extra):
            """Calculate the probs of the next output and update state."""
            h_i, dec_out, context, src_mask = extra
            last = paths[:, :, -1].view(1, -1)
            dec_out, h_i = self.parent.decode_rnn(context, h_i, dec_out, last, src_mask)
            probs = self.parent.output(dec_out)
            dec_out = dec_out.squeeze(0)
            return probs, (h_i, dec_out, context, src_mask)

        def update(self, beams, extra):
            """Select the correct hidden states and outputs to used based on the best performing beams."""
            h_i, dec_out, context, src_mask = extra
            h_i = tuple(hc[:, beams, :] for hc in h_i)
            dec_out = dec_out[beams, :]
            return h_i, dec_out, context, src_mask

    def beam_search(self, encoder_outputs, **kwargs):
        alpha = kwargs.get('alpha')
        if alpha is not None:
            kwargs['length_penalty'] = partial(gnmt_length_penalty, alpha=alpha)
        return RNNDecoder.BeamSearch(parent=self, **kwargs)(encoder_outputs)

    def _greedy_search(self, encoder_output, **kwargs):
        """Decode a sentence by taking the hightest scoring token at each timestep.

        In the past we have just used a beam size of 1 instead of a greedy search because
        they took about the same time to run. I have added this function back because it
        is easier to debug and can help finding where different problems in the output are.

        :param encoder_output: `EncoderOutput` The output of the encoder, it should be
            in the batch first format.
        """
        bsz = encoder_output.output.shape[0]
        device = encoder_output.output.device
        mxlen = int(kwargs.get("mxlen", 100))
        with torch.no_grad():
            src_mask = encoder_output.src_mask  # [B, T]
            # h_i = Tuple[[B, H], [B, H]]
            # dec_out = [B, H]
            # context = [B, T, H]
            h_i, dec_out, context = self.arc_policy(encoder_output, self.hsz)
            # The internal `decode_rnn` actually takes time first so to that.
            last = torch.full((1, bsz), Offsets.GO, dtype=torch.long, device=device)
            outputs = [last]
            for i in range(mxlen - 1):
                # Take a step with the RNN
                # dec_out = [1, B, H]
                # hi = Tuple[[B, H], [B, H]]
                dec_out, h_i = self.decode_rnn(context, h_i, dec_out, last, src_mask) # [1, B, H]
                # Project to vocab size
                probs = self.output(dec_out)  # [1, B, V]
                # Convert the last step of the decorder output into a format we can consume [B, H]
                dec_out = dec_out.squeeze(0)
                # Get the best scoring token for each timestep in the batch
                selected = torch.argmax(probs, dim=-1)
                outputs.append(selected)
                last = selected
            # Combine all the [1, B] outputs into a [T, B] matrix
            outputs = torch.cat(outputs, dim=0)
            # Convert to [B, T]
            outputs = outputs.transpose(0, 1).contiguous()
            # Add a fake beam dimension of size 1
            outputs = outputs.unsqueeze(1)
            # This is mostly for testing so just return zero for lengths and scores.
            return outputs, torch.zeros(bsz), torch.zeros(bsz)


@register_decoder(name='default')
class RNNDecoderWithAttn(RNNDecoder):

    def __init__(self, tgt_embeddings, **kwargs):
        super().__init__(tgt_embeddings, **kwargs)

    def init_attn(self, **kwargs):
        attn_type = kwargs.get('attn_type', 'bahdanau').lower()
        if attn_type == 'dot':
            self.attn_module = LuongDotProductAttention(self.hsz)
        elif attn_type == 'concat' or attn_type == 'bahdanau':
            self.attn_module = BahdanauAttention(self.hsz)
        elif attn_type == 'sdp':
            self.attn_module = ScaledDotProductAttention(self.hsz)
        else:
            self.attn_module = LuongGeneralAttention(self.hsz)

    def attn(self, output_t, context, src_mask=None):
        return self.attn_module(output_t, context, context, src_mask)


@register_decoder(name='transformer')
class TransformerDecoderWrapper(torch.nn.Module):

    def __init__(self, tgt_embeddings, dropout=0.5, layers=1, hsz=None, num_heads=4, scale=True, **kwargs):
        super().__init__()
        self.tgt_embeddings = tgt_embeddings
        dsz = self.tgt_embeddings.get_dsz()
        if hsz is None:
            hsz = dsz

        d_ff = int(kwargs.get('d_ff', 4 * hsz))
        rpr_k = kwargs.get('rpr_k')
        d_k = kwargs.get('d_k')
        activation = kwargs.get('activation', 'relu')
        layer_drop = float(kwargs.get('layer_drop', 0.0))
        scale = bool(kwargs.get('scale', True))

        self.transformer_decoder = TransformerDecoderStack(num_heads, d_model=hsz, d_ff=d_ff,
                                                           pdrop=dropout, scale=scale,
                                                           layers=layers, rpr_k=rpr_k, d_k=d_k,
                                                           activation_type=activation,
                                                           layer_drop=layer_drop)

        self.proj_to_dsz = self._identity
        self.proj_to_hsz = self._identity
        if hsz != dsz:
            self.proj_to_hsz = pytorch_linear(dsz, hsz)
            self.proj_to_dsz = pytorch_linear(hsz, dsz)
            del self.proj_to_dsz.weight
            self.proj_to_dsz.weight = torch.nn.Parameter(self.proj_to_hsz.weight.transpose(0, 1), requires_grad=True)

        do_weight_tying = bool(kwargs.get('tie_weights', True))
        if do_weight_tying:
            if hsz != self.tgt_embeddings.get_dsz():
                raise ValueError("weight tying requires hsz == embedding dsz, got {} hsz and {} dsz".format(self.hsz, self.tgt_embeddings.get_dsz()))
            self.preds = WeightTieDense(self.tgt_embeddings)
        else:
            self.preds = pytorch_linear(dsz, self.tgt_embeddings.get_vsz())

    def _identity(self, x):
        return x

    def forward(self, encoder_output, dst):
        embed_out_bth = self.tgt_embeddings(dst)
        embed_out_bth = self.proj_to_hsz(embed_out_bth)
        context_bth = encoder_output.output
        T = embed_out_bth.shape[1]
        dst_mask = subsequent_mask(T).type_as(embed_out_bth)  # [B, 1, T_q, T_q]
        src_mask = encoder_output.src_mask.unsqueeze(1).unsqueeze(1)  # [B, 1, 1, T_k]
        output = self.transformer_decoder((embed_out_bth, context_bth, src_mask, dst_mask))
        output = self.proj_to_dsz(output)
        prob = self.output(output)
        return prob

    def output(self, x):
        pred = F.log_softmax(self.preds(x.view(x.size(0)*x.size(1), -1)), dim=-1)
        pred = pred.view(x.size(0), x.size(1), -1)
        return pred

    class BeamSearch(BeamSearchBase):

        def __init__(self, parent, **kwargs):
            super().__init__(**kwargs)
            self.parent = parent

        def init(self, encoder_outputs):
            """Tile for the batch of the encoder inputs."""
            encoder_outputs = TransformerEncoderOutput(
                repeat_batch(encoder_outputs.output, self.K),
                repeat_batch(encoder_outputs.src_mask, self.K)
            )
            return encoder_outputs

        def step(self, paths, extra):
            """Calculate the probs for the last item based on the full path."""
            B, K, T = paths.size()
            assert K == self.K
            return self.parent(extra, paths.view(B * K, T))[:, -1], extra

        def update(self, beams, extra):
            """There is no state for the transformer so just pass it."""
            return extra

    def beam_search(self, encoder_outputs, **kwargs):
        return TransformerDecoderWrapper.BeamSearch(parent=self, **kwargs)(encoder_outputs)



import argparse
from baseline.utils import read_config_stream
from mead.utils import hash_config, convert_path


def main():
    parser = argparse.ArgumentParser(description="Get the mead hash of a config.")
    parser.add_argument('config', help='JSON/YML Configuration for an experiment: local file or remote URL', type=convert_path, default="$MEAD_CONFIG")
    args = parser.parse_args()

    config = read_config_stream(args.config)
    print(hash_config(config))


if __name__ == "__main__":
    main()



import os
import re
from typing import List, Dict
import codecs
import json
from itertools import chain
import logging
from collections import Counter
import numpy as np
import baseline.data
from baseline.vectorizers import Dict1DVectorizer, Token1DVectorizer, create_vectorizer, HasPredefinedVocab
from baseline.utils import import_user_module, revlut, exporter, optional_params, Offsets, listify, SingleFileDownloader
from eight_mile.progress import create_progress_bar
__all__ = []
export = exporter(__all__)

logger = logging.getLogger('baseline')


BASELINE_READERS = {}


@export
@optional_params
def register_reader(cls, task, name=None):
    """Register your own `Reader`

    Use this pattern if you want to provide an override to a `Reader` class.

    """
    if name is None:
        name = cls.__name__

    if task not in BASELINE_READERS:
        BASELINE_READERS[task] = {}

    if name in BASELINE_READERS[task]:
        raise Exception('Error: attempt to re-defined previously registered handler {} for task {} in registry'.format(name, task))

    BASELINE_READERS[task][name] = cls
    return cls


@export
def create_reader(task, vectorizers, trim, **kwargs):
    name = kwargs.get('type', kwargs.get('reader_type', 'default'))
    Constructor = BASELINE_READERS[task][name]
    return Constructor(vectorizers, trim, **kwargs)


@export
def num_lines(filename):
    """Counts the number of lines in a file.

    :param filename: `str` The name of the file to count the lines of.
    :returns: `int` The number of lines.
    """
    with codecs.open(filename, encoding='utf-8', mode='r') as f:
        return sum(1 for _ in f)


def _all_predefined_vocabs(vectorizers):
    return all(isinstance(v, HasPredefinedVocab) for v in vectorizers.values())


def _filter_vocab(vocab, min_fs):
    """Filter down the vocab based on rules in the vectorizers.

    :param vocab: `dict[Counter]`: A dict of vocabs.
    :param min_fs: `dict[int]: A dict of cutoffs.

    Note:
        Any key in the min_fs dict should appear in the vocab dict.

    :returns: `dict[dict]`: A dict of new filtered vocabs.
    """
    for k, min_f in min_fs.items():
        # If we don't filter then skip to save an iteration through the vocab
        if min_f == -1:
            continue
        vocab[k] = dict(filter(lambda x: x[1] >= min_f, vocab[k].items()))
    return vocab


def _read_from_col(col, files, col_splitter=r'\t', word_splitter=r'\s'):
    """Read from a single column of a file.

    :param col: `int`: The column to read from.
    :param files: List[str]: A list of files to read from.
    :param col_splitter: `str`: The regex that splits a line into columns.
    :param word_splitter: `str`: The regex that will split a column into words.

    :returns: List[str]: The text from the col of each file.
    """
    text = []
    for file_name in files:
        if file_name is None:
            continue
        with codecs.open(file_name, encoding='utf-8', mode='r') as f:
            for line in f:
                line = line.rstrip('\n')
                if line == "":
                    continue
                cols = re.split(col_splitter, line)
                text.append(re.split(word_splitter, cols[col]))
    return text


def _build_vocab_for_col(col, files, vectorizers, text=None, col_splitter=r'\t', word_splitter=r'\s'):
    """Build vocab from a single column in file. (separated by `\t`).

    Used to read a vocab from a single conll column, read a vocab from the
    source or target of a seq2seq file, or reading from a vocab file.

    :param col: `int`: The column to read from.
    :param files: List[str]: A list of files to read from.
    :param vectorizers: dict[str] -> Vectorizer: The vectorizer to use to count the column
    :param text: List[str]: The text from the columns or None
    :param col_splitter: `str`: The regex that splits a line into columns.
    :param word_splitter: `str`: The regex that will split a column into words.

    :returns: dict[str] -> dict[str] -> int: The vocabs.
    """
    text = _read_from_col(col, files, col_splitter, word_splitter) if text is None else text
    vocab = {k: Counter() for k in vectorizers}
    for t in text:
        for k, vect in vectorizers.items():
            vocab[k].update(vect.count(t))
    return vocab


def _check_lens(vectorizers):
    failures = set()
    for k, vect in vectorizers.items():
        mxlen = getattr(vect, 'mxlen', None)
        if mxlen == -1:
            failures.add(k)
    return failures


def _vocab_allowed(vectorizers):
    fails = _check_lens(vectorizers)
    if fails:
        fail_str = "When using a vocab file mxlen for vectorizers must not be `-1`\n"
        vect_str = "\n".join("\t{}".format(fails))
        raise RuntimeError(fail_str + vect_str)


@export
class ParallelCorpusReader:

    def __init__(self, vectorizers, trim=False, truncate=False):
        super().__init__()

        self.src_vectorizers = {}
        self.tgt_vectorizer = None
        for k, vectorizer in vectorizers.items():
            if k == 'tgt':
                self.tgt_vectorizer = vectorizer
                if not self.tgt_vectorizer.emit_begin_tok:
                    self.tgt_vectorizer.emit_begin_tok.append(Offsets.VALUES[Offsets.GO])
                if not self.tgt_vectorizer.emit_end_tok:
                    self.tgt_vectorizer.emit_end_tok.append(Offsets.VALUES[Offsets.EOS])
            else:
                self.src_vectorizers[k] = vectorizer
        self.trim = trim
        self.truncate = truncate

    def build_vocabs(self, files, **kwargs):
        pass

    def load_examples(self, tsfile, vocab1, vocab2, shuffle, sort_key):
        pass

    def load(self, tsfile, vocab1, vocab2, batchsz, shuffle=False, sort_key=None):
        examples = self.load_examples(tsfile, vocab1, vocab2, shuffle, sort_key)
        return baseline.data.ExampleDataFeed(examples, batchsz,
                                             shuffle=shuffle, trim=self.trim, sort_key=sort_key, truncate=self.truncate)


@register_reader(task='seq2seq', name='tsv')
class TSVParallelCorpusReader(ParallelCorpusReader):

    def __init__(self, vectorizers,
                 trim=False, truncate=False, src_col_num=0, tgt_col_num=1, **kwargs):
        super().__init__(vectorizers, trim, truncate)
        self.src_col_num = src_col_num
        self.tgt_col_num = tgt_col_num

    def build_vocabs(self, files, **kwargs):
        if _all_predefined_vocabs(self.src_vectorizers) and isinstance(self.tgt_vectorizer, HasPredefinedVocab):
            logger.info("Skipping building vocabulary.  All vectorizers have predefined vocabs!")
            return {k: v.vocab for k, v in self.src_vectorizers.items()}, self.tgt_vectorizer.vocab
        vocab_file = kwargs.get('vocab_file')
        if vocab_file is not None:
            all_vects = self.src_vectorizers.copy()
            all_vects['tgt'] = self.tgt_vectorizer
            _vocab_allowed(all_vects)
            # Only read the file once
            text = _read_from_col(0, listify(vocab_file))
            src_vocab = _build_vocab_for_col(None, None, self.src_vectorizers, text=text)
            tgt_vocab = _build_vocab_for_col(None, None, {'tgt': self.tgt_vectorizer}, text=text)
            return src_vocab, tgt_vocab['tgt']
        src_vocab = _build_vocab_for_col(self.src_col_num, files, self.src_vectorizers)
        tgt_vocab = _build_vocab_for_col(self.tgt_col_num, files, {'tgt': self.tgt_vectorizer})
        min_f = kwargs.get('min_f', {})
        tgt_min_f = {'tgt': min_f.pop('tgt', -1)}
        src_vocab = _filter_vocab(src_vocab, min_f)
        tgt_vocab = _filter_vocab(tgt_vocab, tgt_min_f)
        return src_vocab, tgt_vocab['tgt']

    def load_examples(self, tsfile, src_vocabs, tgt_vocab, do_shuffle, src_sort_key):
        ts = []
        with codecs.open(tsfile, encoding='utf-8', mode='r') as f:
            for line in f:
                splits = re.split("\t", line.strip())
                src = list(filter(lambda x: len(x) != 0, re.split("\s+", splits[0])))

                example = {}
                for k, vectorizer in self.src_vectorizers.items():
                    example[k], length = vectorizer.run(src, src_vocabs[k])
                    if length is not None:
                        example['{}_lengths'.format(k)] = length

                tgt = list(filter(lambda x: len(x) != 0, re.split("\s+", splits[1])))
                example['tgt'], example['tgt_lengths'] = self.tgt_vectorizer.run(tgt, tgt_vocab)
                ts.append(example)
        return baseline.data.Seq2SeqExamples(ts, do_shuffle=do_shuffle, src_sort_key=src_sort_key)


@export
@register_reader(task='seq2seq', name='default')
class MultiFileParallelCorpusReader(ParallelCorpusReader):

    def __init__(self, vectorizers, trim=False, truncate=False, **kwargs):
        super().__init__(vectorizers, trim, truncate)
        pair_suffix = kwargs['pair_suffix']

        self.src_suffix = pair_suffix[0]
        self.tgt_suffix = pair_suffix[1]
        if not self.src_suffix.startswith('.'):
            self.src_suffix = '.' + self.src_suffix
        if not self.tgt_suffix.startswith('.'):
            self.tgt_suffix = '.' + self.tgt_suffix

    def build_vocabs(self, files, **kwargs):
        if _all_predefined_vocabs(self.src_vectorizers) and isinstance(self.tgt_vectorizer, HasPredefinedVocab):
            logger.info("Skipping building vocabulary.  All vectorizers have predefined vocabs!")
            return {k: v.vocab for k, v in self.src_vectorizers.items()}, self.tgt_vectorizer.vocab
        vocab_file = kwargs.get('vocab_file')
        if vocab_file is not None:
            all_vects = self.src_vectorizers.copy()
            all_vects['tgt'] = self.tgt_vectorizer
            _vocab_allowed(all_vects)
            # Only read the file once.
            text = _read_from_col(0, listify(vocab_file))
            src_vocab = _build_vocab_for_col(None, None, self.src_vectorizers, text=text)
            tgt_vocab = _build_vocab_for_col(None, None, {'tgt': self.tgt_vectorizer}, text=text)
            return src_vocab, tgt_vocab['tgt']
        src_vocab = _build_vocab_for_col(0, [f + self.src_suffix for f in files], self.src_vectorizers)
        tgt_vocab = _build_vocab_for_col(0, [f + self.tgt_suffix for f in files], {'tgt': self.tgt_vectorizer})
        min_f = kwargs.get('min_f', {})
        tgt_min_f = {'tgt': min_f.pop('tgt', -1)}
        src_vocab = _filter_vocab(src_vocab, min_f)
        tgt_vocab = _filter_vocab(tgt_vocab, tgt_min_f)
        return src_vocab, tgt_vocab['tgt']

    def load_examples(self, tsfile, src_vocabs, tgt_vocab, do_shuffle, src_sort_key):
        ts = []

        with codecs.open(tsfile + self.src_suffix, encoding='utf-8', mode='r') as fsrc:
            with codecs.open(tsfile + self.tgt_suffix, encoding='utf-8', mode='r') as ftgt:
                for src, tgt in zip(fsrc, ftgt):
                    example = {}
                    src = re.split("\s+", src.strip())
                    for k, vectorizer in self.src_vectorizers.items():
                        example[k], length = vectorizer.run(src, src_vocabs[k])
                        if length is not None:
                            example['{}_lengths'.format(k)] = length
                    tgt = re.split("\s+", tgt.strip())
                    example['tgt'], example['tgt_lengths'] = self.tgt_vectorizer.run(tgt, tgt_vocab)
                    ts.append(example)
        return baseline.data.Seq2SeqExamples(ts, do_shuffle=do_shuffle, src_sort_key=src_sort_key)


def _try_read_labels(**kwargs):
    label_file = kwargs.get('label_file')
    label_list = kwargs.get('label_list')
    label2index = {}
    if label_file:
        pre_labels = Counter(chain(*_read_from_col(0, listify(label_file))))
        label2index = {l: i for i, l in enumerate(pre_labels)}
    elif label_list:
        label2index = {l: i for i, l in enumerate(label_list)}
    return label2index

@export
class SeqPredictReader:

    def __init__(self, vectorizers, trim=False, truncate=False, mxlen=-1, **kwargs):
        super().__init__()
        self.vectorizers = vectorizers
        self.trim = trim
        self.truncate = truncate
        label_vectorizer_spec = kwargs.get('label_vectorizer', None)
        if label_vectorizer_spec:
            cache = label_vectorizer_spec.get("data_download_cache", os.path.expanduser("~/.bl-data"))
            if 'model_file' in label_vectorizer_spec:
                label_vectorizer_spec['model_file'] = SingleFileDownloader(label_vectorizer_spec['model_file'], cache).download()
            if 'vocab_file' in label_vectorizer_spec:
                label_vectorizer_spec['vocab_file'] = SingleFileDownloader(label_vectorizer_spec['vocab_file'], cache).download()
            if 'transform' in label_vectorizer_spec:
                label_vectorizer_spec['transform_fn'] = label_vectorizer_spec['transform']
            if 'transform_fn' in label_vectorizer_spec and isinstance(label_vectorizer_spec['transform_fn'], str):
                label_vectorizer_spec['transform_fn'] = eval(label_vectorizer_spec['transform_fn'])
            self.label_vectorizer = create_vectorizer(**label_vectorizer_spec)
        else:
            self.label_vectorizer = Dict1DVectorizer(fields='y', mxlen=mxlen)
        self.label2index = {
            Offsets.VALUES[Offsets.PAD]: Offsets.PAD,
            Offsets.VALUES[Offsets.GO]: Offsets.GO,
            Offsets.VALUES[Offsets.EOS]: Offsets.EOS
        }

    def build_vocab(self, files, **kwargs):

        # TODO: Im not sure we should even support this option
        label2index = _try_read_labels(**kwargs)
        if label2index and _all_predefined_vocabs(self.vectorizers):
            offset = len(label2index)
            self.label2index.update({k: v + offset} for k, v in label2index.items())
            logger.info("Skipping building vocabulary.  All vectorizers have predefined vocabs and a label file was given!")
            return {k: v.vocab for k, v in self.vectorizers.items()}
        pre_vocabs = None
        vocabs = {k: Counter() for k in self.vectorizers.keys()}

        vocab_file = kwargs.get('vocab_file')

        if vocab_file:
            _vocab_allowed(self.vectorizers)
            pre_vocabs = _build_vocab_for_col(0, listify(vocab_file), self.vectorizers)

        labels = Counter()

        for file in files:
            if file is None:
                continue

            examples = self.read_examples(file)
            for example in examples:
                labels.update(self.label_vectorizer.count(example))
                for k, vectorizer in self.vectorizers.items():
                    vocab_example = vectorizer.count(example)
                    vocabs[k].update(vocab_example)

        if label2index and not pre_vocabs:
            return vocabs

        vocabs = _filter_vocab(vocabs, kwargs.get('min_f', {}))
        base_offset = len(self.label2index)
        labels.pop(Offsets.VALUES[Offsets.PAD], None)
        for i, k in enumerate(labels.keys()):
            self.label2index[k] = i + base_offset
        if pre_vocabs:
            vocabs = pre_vocabs
        return vocabs

    def read_examples(self):
        pass

    def load(self, filename, vocabs, batchsz, shuffle=False, sort_key=None):

        ts = []
        texts = self.read_examples(filename)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'

        for i, example_tokens in enumerate(texts):
            example = {}
            for k, vectorizer in self.vectorizers.items():
                example[k], lengths = vectorizer.run(example_tokens, vocabs[k])
                if lengths is not None:
                    example['{}_lengths'.format(k)] = lengths
            example['y'], lengths = self.label_vectorizer.run(example_tokens, self.label2index)
            example['y_lengths'] = lengths
            example['ids'] = i
            ts.append(example)

        examples = baseline.data.DictExamples(ts, do_shuffle=shuffle, sort_key=sort_key)
        return baseline.data.ExampleDataFeed(examples, batchsz=batchsz, shuffle=shuffle, trim=self.trim, truncate=self.truncate), texts


@export
class MultiLabelSeqPredictReader:

    def __init__(self, vectorizers, trim=False, truncate=False, mxlen=-1, **kwargs):
        super().__init__()
        self.vectorizers = vectorizers
        self.trim = trim
        self.truncate = truncate
        label_vectorizer_spec_dict = kwargs.get('label_vectorizers', {'y': Dict1DVectorizer(fields='y', mxlen=mxlen)})
        self.label_vectorizers = {}
        self.label2index = {}
        for k, label_vectorizer_spec in label_vectorizer_spec_dict.items():
            if 'label' not in label_vectorizer_spec:
                label_vectorizer_spec['label'] = k
            cache = label_vectorizer_spec.get("data_download_cache", os.path.expanduser("~/.bl-data"))
            if 'model_file' in label_vectorizer_spec:
                label_vectorizer_spec['model_file'] = SingleFileDownloader(label_vectorizer_spec['model_file'], cache).download()
            if 'vocab_file' in label_vectorizer_spec:
                label_vectorizer_spec['vocab_file'] = SingleFileDownloader(label_vectorizer_spec['vocab_file'], cache).download()
            label_vectorizer = create_vectorizer(**label_vectorizer_spec)

            self.label_vectorizers[k] = label_vectorizer
            self.label2index[k] = {
                Offsets.VALUES[Offsets.PAD]: Offsets.PAD,
                Offsets.VALUES[Offsets.GO]: Offsets.GO,
                Offsets.VALUES[Offsets.EOS]: Offsets.EOS
            }

    def build_vocab(self, files, **kwargs):
        """For this model, we are not going to support label pre-definitions

        :param files:
        :param kwargs:
        :return:
        """
        pre_vocabs = None
        vocabs = {k: Counter() for k in self.vectorizers.keys()}
        labels = {k: Counter() for k in self.label_vectorizers.keys()}

        vocab_file = kwargs.get('vocab_file')

        if vocab_file:
            _vocab_allowed(self.vectorizers)
            pre_vocabs = _build_vocab_for_col(0, listify(vocab_file), self.vectorizers)

        for file in files:
            if file is None:
                continue

            examples = self.read_examples(file)
            for example in examples:
                for k, vectorizer in self.label_vectorizers.items():
                    labels[k].update(vectorizer.count(example))
                for k, vectorizer in self.vectorizers.items():
                    vocab_example = vectorizer.count(example)
                    vocabs[k].update(vocab_example)

        vocabs = _filter_vocab(vocabs, kwargs.get('min_f', {}))
        base_offset = len(Offsets.VALUES) - 1  # Dont put UNK in labels

        for l in labels.keys():
            labels[l].pop(Offsets.VALUES[Offsets.PAD], None)
            for i, k in enumerate(labels[l]):
                self.label2index[l][k] = i + base_offset
        if pre_vocabs:
            vocabs = pre_vocabs
        return vocabs

    def read_examples(self):
        pass

    def load(self, filename, vocabs, batchsz, shuffle=False, sort_key=None):

        ts = []
        texts = self.read_examples(filename)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'

        for i, example_tokens in enumerate(texts):
            example = {}
            for k, vectorizer in self.vectorizers.items():
                example[k], lengths = vectorizer.run(example_tokens, vocabs[k])
                if lengths is not None:
                    example['{}_lengths'.format(k)] = lengths
            for k, vectorizer in self.label_vectorizers.items():
                example[k], lengths = vectorizer.run(example_tokens, self.label2index[k])
                if lengths is not None:
                    example['{}_lengths'.format(k)] = lengths
            example['ids'] = i
            if np.max(example['heads']) >= example['word_lengths']:
                raise Exception("Invalid head", np.max(example['heads']), example['word_lengths'])
            ts.append(example)

        examples = baseline.data.DictExamples(ts, do_shuffle=shuffle, sort_key=sort_key)
        return baseline.data.ExampleDataFeed(examples, batchsz=batchsz, shuffle=shuffle, trim=self.trim, truncate=self.truncate), texts


@export
@register_reader(task='tagger', name='default')
class CONLLSeqReader(SeqPredictReader):

    def __init__(self, vectorizers, trim=False, truncate=False, mxlen=-1, **kwargs):
        super().__init__(vectorizers, trim, truncate, mxlen, **kwargs)
        self.named_fields = kwargs.get('named_fields', {})

    def read_examples(self, tsfile):

        tokens = []
        examples = []

        with codecs.open(tsfile, encoding='utf-8', mode='r') as f:
            for i, line in enumerate(f):
                states = re.split("\s", line.strip())

                token = dict()
                if len(states) > 1:
                    for j in range(len(states)):
                        noff = j - len(states)
                        if noff >= 0:
                            noff = j
                        field_name = self.named_fields.get(str(j),
                                                           self.named_fields.get(str(noff), str(j)))
                        token[field_name] = states[j]
                    tokens.append(token)

                else:
                    if len(tokens) == 0:
                        raise Exception("Unexpected empty line ({}) in {}".format(i, tsfile))
                    examples.append(tokens)
                    tokens = []
            if len(tokens) > 0:
                examples.append(tokens)
        return examples



@export
@register_reader(task='deps', name='default')
class CONLLParserSeqReader(MultiLabelSeqPredictReader):

    def __init__(self, vectorizers, trim=False, truncate=False, mxlen=-1, **kwargs):
        super().__init__(vectorizers, trim, truncate, mxlen, **kwargs)
        self.named_fields = kwargs.get('named_fields', {})

    def read_examples(self, tsfile):

        tokens = []
        examples = []

        with codecs.open(tsfile, encoding='utf-8', mode='r') as f:
            for i, line in enumerate(f):
                states = re.split("\s", line.strip())

                token = dict()
                if len(states) > 1:
                    for j in range(len(states)):
                        noff = j - len(states)
                        if noff >= 0:
                            noff = j
                        field_name = self.named_fields.get(str(j),
                                                           self.named_fields.get(str(noff), str(j)))
                        token[field_name] = states[j]
                    tokens.append(token)

                else:
                    if len(tokens) == 0:
                        raise Exception("Unexpected empty line ({}) in {}".format(i, tsfile))
                    examples.append(tokens)
                    tokens = []
            if len(tokens) > 0:
                examples.append(tokens)
        return examples


@export
@register_reader(task='tagger', name='srl')
class SRLSeqReader(SeqPredictReader):
    """Read data in the format described by https://github.com/luheng/deep_srl tested on CoNLL2012 SRL data

    """
    def __init__(self, vectorizers, trim=False, truncate=False, mxlen=-1, **kwargs):
        super().__init__(vectorizers, trim, truncate, mxlen, **kwargs)
        self.named_fields = kwargs.get('named_fields', {})

    def read_examples(self, tsfile):

        examples = []

        with codecs.open(tsfile, encoding='utf-8', mode='r') as f:
            for i, line in enumerate(f):
                pred_surface, labels = line.strip().split("|||")
                pred_surface = pred_surface.strip().split()
                labels = labels.strip().split()
                pred = int(pred_surface[0])
                surface = pred_surface[1:]
                if len(surface) > len(labels):
                    logger.warning("No labels given for line [%s]", line)
                    labels = ['O'] * len(surface)

                tokens = [{'text': s, 'y': l, 'pred': 0} for s, l in zip(surface, labels)]
                tokens[pred]['pred'] = 1
                if tokens:
                    examples.append(tokens)
        return examples


def _norm_ext(ext):
    return ext if ext.startswith('.') else '.' + ext


@export
class SeqLabelReader:

    def __init__(self):
        pass

    def build_vocab(self, files, **kwargs):
        pass

    def load(self, filename, index, batchsz, **kwargs):
        pass


def _get_dir(files):
    if isinstance(files, str):
        if os.path.isdir(files):
            base = files
            files = filter(os.path.isfile, [os.path.join(base, x) for x in os.listdir(base)])
        else:
            files = [files]
    return files


@export
class LineSeqLabelReader(SeqLabelReader):

    REPLACE = { "'s": " 's ",
                "'ve": " 've ",
                "n't": " n't ",
                "'re": " 're ",
                "'d": " 'd ",
                "'ll": " 'll ",
                ",": " , ",
                "!": " ! ",
                }

    def __init__(self, vectorizers, trim=False, truncate=False, **kwargs):
        super().__init__()

        self.label2index = {}
        self.vectorizers = vectorizers
        self.clean_fn = kwargs.get('clean_fn')
        if self.clean_fn is None:
            self.clean_fn = lambda x: x
        self.trim = trim
        self.truncate = truncate
        self.has_header = bool(kwargs.get('has_header', False))
        self.col_keys = kwargs.get('col_keys', [])
    SPLIT_ON = '[\t\s]+'

    @staticmethod
    def splits(text):
        return list(filter(lambda s: len(s) != 0, re.split('\s+', text)))

    @staticmethod
    def do_clean(l):
        l = l.lower()
        l = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", l)
        for k, v in TSVSeqLabelReader.REPLACE.items():
            l = l.replace(k, v)
        return l.strip()

    def label_and_sentence(self, line, clean_fn, header):
        """This function produces the sample from a line

        :param line: The raw line
        :param clean_fn: A clean function
        :param header: An optional header
        :return: A tuple of the label and text sample
        """

    def build_vocab(self, files, **kwargs):
        """Take a directory (as a string), or an array of files and build a vocabulary

        Take in a directory or an array of individual files (as a list).  If the argument is
        a string, it may be a directory, in which case, all files in the directory will be loaded
        to form a vocabulary.

        :param files: Either a directory (str), or an array of individual files
        :return:
        """
        vocab_file = kwargs.get('vocab_file')
        self.label2index = _try_read_labels(**kwargs)

        if _all_predefined_vocabs(self.vectorizers):
            if not self.label2index:
                logger.warning("If you provide a label file or list, we can skip tabulating the labels from the data!")
                label_idx = 0
                for file_name in files:
                    with codecs.open(file_name, encoding='utf-8', mode='r') as f:
                        header = next(f) if self.has_header else None
                        for il, line in enumerate(f):
                            label, _ = self.label_and_sentence(line, self.clean_fn, header)
                            if label not in self.label2index:
                                self.label2index[label] = label_idx
                                label_idx += 1
            else:
                logger.info("Skipping label file and vocabulary!")
            logger.info("Skipping building vocabulary!")
            return {k: v.vocab for k, v in self.vectorizers.items()}, self.get_labels()

        if vocab_file and self.label2index:
            _vocab_allowed(self.vectorizers)
            vocab = _build_vocab_for_col(0, listify(vocab_file), self.vectorizers)
            return vocab, self.get_labels()

        label_idx = len(self.label2index)
        if isinstance(files, str):
            if os.path.isdir(files):
                base = files
                files = filter(os.path.isfile, [os.path.join(base, x) for x in os.listdir(base)])
            else:
                files = [files]
        vocab = {k: Counter() for k in self.vectorizers.keys()}

        for file_name in files:
            if file_name is None:
                continue
            with codecs.open(file_name, encoding='utf-8', mode='r') as f:
                header = next(f) if self.has_header else None
                for il, line in enumerate(f):
                    label, text = self.label_and_sentence(line, self.clean_fn, header)
                    if len(text) == 0:
                        continue

                    for k, vectorizer in self.vectorizers.items():
                        vocab_file = vectorizer.count(text)
                        vocab[k].update(vocab_file)

                    if label not in self.label2index:
                        self.label2index[label] = label_idx
                        label_idx += 1

        vocab = _filter_vocab(vocab, kwargs.get('min_f', {}))

        return vocab, self.get_labels()

    def get_labels(self):
        labels = [''] * len(self.label2index)
        for label, index in self.label2index.items():
            labels[index] = label
        return labels

    def load_text(self, filename, vocabs, batchsz, **kwargs):

        shuffle = kwargs.get('shuffle', False)
        sort_key = kwargs.get('sort_key', None)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'

        examples = []
        texts = []
        with codecs.open(filename, encoding='utf-8', mode='r') as f:
            header = next(f) if self.has_header else None
            for il, line in enumerate(f):
                label, text = self.label_and_sentence(line, self.clean_fn, header)
                texts.append(text)
                if len(text) == 0:
                    continue
                if label not in self.label2index:
                    continue
                y = self.label2index[label]
                example_dict = dict()
                for k, vectorizer in self.vectorizers.items():
                    example_dict[k], lengths = vectorizer.run(text, vocabs[k])
                    if lengths is not None:
                        example_dict['{}_lengths'.format(k)] = lengths

                example_dict['y'] = y
                examples.append(example_dict)
        return baseline.data.ExampleDataFeed(baseline.data.DictExamples(examples,
                                                                        do_shuffle=shuffle,
                                                                        sort_key=sort_key),
                                             batchsz=batchsz, shuffle=shuffle, trim=self.trim, truncate=self.truncate), texts

    def load(self, filename, vocabs, batchsz, **kwargs):
    
        shuffle = kwargs.get('shuffle', False)
        sort_key = kwargs.get('sort_key', None)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'
    
        examples = []

        with codecs.open(filename, encoding='utf-8', mode='r') as f:
            header = next(f) if self.has_header else None
            for il, line in enumerate(f):
                label, text = self.label_and_sentence(line, self.clean_fn, header)
                if len(text) == 0:
                    continue
                y = self.label2index[label]
                example_dict = dict()
                for k, vectorizer in self.vectorizers.items():
                    example_dict[k], lengths = vectorizer.run(text, vocabs[k])
                    if lengths is not None:
                        example_dict['{}_lengths'.format(k)] = lengths
            
                example_dict['y'] = y
                examples.append(example_dict)
        return baseline.data.ExampleDataFeed(baseline.data.DictExamples(examples,
                                                                        do_shuffle=shuffle,
                                                                        sort_key=sort_key),
                                             batchsz=batchsz, shuffle=shuffle, trim=self.trim, truncate=self.truncate)


@export
@register_reader(task='classify', name='default')
class TSVSeqLabelReader(LineSeqLabelReader):

    def label_and_sentence(self, line, clean_fn, header):
        if header is None:
            return self.label_and_sentence_no_header(line, clean_fn)
        cols = line.strip().split('\t')
        header = [h.strip() for h in header.split('\t')]
        col_indices = [header.index(c) for c in self.col_keys]
        label_sentence = [cols[c] for c in col_indices]
        label, sentence = label_sentence
        text = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in sentence.split()])))
        return label, text

    def label_and_sentence_no_header(self, line, clean_fn):
        label_text = re.split(TSVSeqLabelReader.SPLIT_ON, line)
        label = label_text[0]
        text = label_text[1:]
        text = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in text])))
        text = TSVSeqLabelReader.splits(text)
        return label, text


class IndexPairLabelReader(SeqLabelReader):
    """Shared vocabulary paired label reader

    This is a TSV reader like TSVSeqLabelReader, but the vocabulary and vectorizer are used to produce a paired input.
    This would be useful, for example, if the problem is e.g. NLI classification.
    In this implementation, we assume that the user wishes to share the vocab and vectorizer for the 2 inputs, and
    we use them to generate a key that is slightly different from the user key defined.  So for instance, if the user
    defines the key as `x`, in the `TSVSeqLabelReader` this would produce a vector `x` and `x_lengths`.  In this reader,
    however, the keys will be generated as `x[0]` and `x[0]_lengths` for the first item in the pair, and index `1` for
    the second
    """
    REPLACE = {"'s": " 's ",
               "'ve": " 've ",
               "n't": " n't ",
               "'re": " 're ",
               "'d": " 'd ",
               "'ll": " 'll ",
               ",": " , ",
               "!": " ! ",
               }

    def __init__(self, vectorizers, trim=False, truncate=False, **kwargs):
        super().__init__()

        self.label2index = {}
        self.vectorizers = vectorizers
        self.clean_fn = kwargs.get('clean_fn')
        if self.clean_fn is None:
            self.clean_fn = lambda x: x
        self.trim = trim
        self.truncate = truncate
        self.use_token_type = bool(kwargs.get('use_token_type', False))
        self.start_tokens_1 = kwargs.get('start_tokens_1', [])
        self.start_tokens_2 = kwargs.get('start_tokens_2', [])
        self.end_tokens_1 = kwargs.get('end_tokens_1', [])
        self.end_tokens_2 = kwargs.get('end_tokens_2', [])

        self.col_keys = kwargs.get('col_keys', ['pairID', 'sentence1', 'sentence2', 'gold_label'])
        example_type = kwargs.get('example_type', 'single')
        self.convert_to_example = self.convert_to_example_dual if example_type == 'dual' else self.convert_to_example_single
        self.has_header = True

    @staticmethod
    def splits(text):
        return list(filter(lambda s: len(s) != 0, re.split('\s+', text)))

    @staticmethod
    def do_clean(l):
        l = l.lower()
        l = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", l)
        for k, v in IndexPairLabelReader.REPLACE.items():
            l = l.replace(k, v)
        return l.strip()

    def index_pair_label(self, line, clean_fn, header):
        """Does the actual work of handling the file format abstractions.

        :param line: A line of text fromn the file
        :param clean_fn: A cleaning function to apply
        :return: A tuple of the shape (index, text1, text2, label)
        """

    def build_vocab(self, files, **kwargs):
        """Take a directory (as a string), or an array of files and build a vocabulary

        Take in a directory or an array of individual files (as a list).  If the argument is
        a string, it may be a directory, in which case, all files in the directory will be loaded
        to form a vocabulary.

        :param files: Either a directory (str), or an array of individual files
        :return:
        """
        vocab_file = kwargs.get('vocab_file')
        self.label2index = _try_read_labels(**kwargs)
        if _all_predefined_vocabs(self.vectorizers):
            if not self.label2index:
                logger.warning("If you provide a label file, we can skip tabulating the labels from the data!")
                label_idx = 0
                for file_name in files:
                    with codecs.open(file_name, encoding='utf-8', mode='r') as f:
                        header = next(f) if self.has_header else None
                        for il, line in enumerate(f):
                            idx, text1, text2, label = self.index_pair_label(line, self.clean_fn, header)
                            if label not in self.label2index:
                                self.label2index[label] = label_idx
                                label_idx += 1
            else:
                logger.info("Skipping label file and vocabulary!")
            logger.info("Skipping building vocabulary!")
            return {k: v.vocab for k, v in self.vectorizers.items()}, self.get_labels()

        if vocab_file is not None and self.label2index:
            _vocab_allowed(self.vectorizers)
            vocab = _build_vocab_for_col(0, listify(vocab_file), self.vectorizers)
            return vocab, self.get_labels()

        label_idx = len(self.label2index)
        if isinstance(files, str):
            if os.path.isdir(files):
                base = files
                files = filter(os.path.isfile, [os.path.join(base, x) for x in os.listdir(base)])
            else:
                files = [files]
        vocab = {k: Counter() for k in self.vectorizers.keys()}

        for file_name in files:
            if file_name is None:
                continue
            with codecs.open(file_name, encoding='utf-8', mode='r') as f:
                header = next(f) if self.has_header else None
                for il, line in enumerate(f):
                    idx, text1, text2, label = self.index_pair_label(line, self.clean_fn, header)
                    if len(text1) < 1 or len(text2) < 1:
                        raise Exception(f"Invalid line @ {il}: [{line}]")

                    for k, vectorizer in self.vectorizers.items():
                        vocab_file = vectorizer.count(text1) + vectorizer.count(text2)
                        vocab[k].update(vocab_file)

                    if label not in self.label2index:
                        self.label2index[label] = label_idx
                        label_idx += 1
                    #if il > 0 and il % 1000 == 0:
                    #    print(il)
        vocab = _filter_vocab(vocab, kwargs.get('min_f', {}))
        return vocab, self.get_labels()

    def get_labels(self):
        labels = [''] * len(self.label2index)
        for label, index in self.label2index.items():
            labels[index] = label
        return labels

    def load_text(self, filename, vocabs, batchsz, **kwargs):

        shuffle = kwargs.get('shuffle', False)
        sort_key = kwargs.get('sort_key', None)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'

        examples = []
        texts = []
        with codecs.open(filename, encoding='utf-8', mode='r') as f:
            header = next(f) if self.has_header else None
            for il, line in enumerate(f):
                idx, text1, text2, label = self.index_pair_label(line, self.clean_fn, header)
                if len(text1) < 1 or len(text2) < 1:
                    raise Exception(f"Invalid line @ {il}: [{line}]")
                example_dict = self.convert_to_example(vocabs, text1, text2)
                example_dict['y'] = self.label2index[label]
                example_dict['idx'] = idx
                examples.append(example_dict)

        return baseline.data.ExampleDataFeed(baseline.data.DictExamples(examples,
                                                                        do_shuffle=shuffle,
                                                                        sort_key=sort_key),
                                             batchsz=batchsz, shuffle=shuffle, trim=self.trim,
                                             truncate=self.truncate), texts

    def convert_to_example_dual(self, vocabs: Dict[str, str], text1: List[str], text2: List[str]) -> Dict[str, object]:
        """Convert to 2 vectors, Useful for dual-encoder-style models

        :param vocabs: Vocabs to use
        :param text1: First text
        :param text2: Second text
        :return: An example dict
        """
        text = [text1, text2]
        example_dict = dict()
        for pair_idx in range(2):
            for vec_name, vectorizer in self.vectorizers.items():
                example_dict[f"{vec_name}[{pair_idx}]"], lengths = vectorizer.run(text[pair_idx], vocabs[vec_name])
                if lengths is not None:
                    example_dict[f'{vec_name}[{pair_idx}]_lengths'] = lengths
        return example_dict

    def convert_to_example_single(self, vocabs: Dict[str, str], text1: List[str], text2: List[str]) -> Dict[str, object]:
        """Convert to a concatenated vector.  Useful for BERT-style models

        :param vocabs: Vocabs to use
        :param text1: First text
        :param text2: Second text
        :return: An example dict
        """
        example_dict = dict()

        for key, vectorizer in self.vectorizers.items():
            v = vocabs[key]
            vec_1, lengths = vectorizer.run(text1, v)
            vec_1 = vec_1[:lengths]
            vec_2, lengths = vectorizer.run(text2, v)
            vec_2 = vec_2[:lengths]

            if self.start_tokens_1:
                vec_1 = np.concatenate([np.array([v[x] for x in self.start_tokens_1]), vec_1])
            if self.end_tokens_1:
                vec_1 = np.concatenate([vec_1, np.array([v[x] for x in self.end_tokens_1])])

            if self.start_tokens_2:
                vec_2 = np.concatenate([np.array([v[x] for x in self.start_tokens_2]), vec_2])
            if self.end_tokens_2:
                vec_2 = np.concatenate([vec_2, np.array([v[x] for x in self.end_tokens_2])])
            lengths_1 = vec_1.shape[-1]
            lengths_2 = vec_2.shape[-1]
            total_length = lengths_1 + lengths_2
            vec = np.concatenate([vec_1, vec_2])
            extend = vectorizer.mxlen - total_length

            if extend < 0:
                logger.warning(f"total length {total_length} exceeds allocated vectorizer width of {vectorizer.mxlen}")
                vec = vec[:vectorizer.mxlen]
            elif extend > 0:
                vec = np.concatenate([vec, np.zeros(extend, vec.dtype)])

            example_dict[key] = vec
            if self.use_token_type:
                token_type = np.zeros_like(vec)
                token_type[lengths_1:] = 1
                example_dict[f"{key}_tt"] = token_type
        return example_dict

    def load(self, filename, vocabs, batchsz, **kwargs):

        shuffle = kwargs.get('shuffle', False)
        sort_key = kwargs.get('sort_key', None)
        if sort_key is not None and not sort_key.endswith('_lengths'):
            sort_key += '_lengths'

        examples = []
        with open(filename, encoding='utf-8-sig', mode='r') as f:
            header = next(f) if self.has_header else None
            #f = list(f)
            ##pg = create_progress_bar(len(f), 'tqdm')
            for il, line in enumerate(f):
                try:
                    idx, text1, text2, label = self.index_pair_label(line, self.clean_fn, header)
                    if len(text1) < 1 or len(text2) < 1:
                        raise Exception(f"Invalid line @ {il}: [{line}]")
                    example_dict = self.convert_to_example(vocabs, text1.split(), text2.split())
                    if label not in self.label2index:
                        continue
                    example_dict['y'] = self.label2index[label]
                    example_dict['idx'] = idx
                    examples.append(example_dict)
                except Exception as e:
                    logger.warning(f'Skipping invalid line {il}')

        return baseline.data.ExampleDataFeed(baseline.data.DictExamples(examples,
                                                                        do_shuffle=shuffle,
                                                                        sort_key=sort_key),
                                             batchsz=batchsz, shuffle=shuffle, trim=self.trim, truncate=self.truncate)


@export
@register_reader(task='classify', name='tsv-paired-shared-vec')
class TSVIndexPairLabelReader(IndexPairLabelReader):

    def index_pair_label(self, line, clean_fn, header):
        cols = line.strip().split('\t')
        header = [h.strip() for h in header.split('\t')]
        col_indices = [header.index(c) for c in self.col_keys]
        # QQP files have some invalid samples
        #if max(col_indices) > len(cols):
        #    raise Exception(f"Error: {cols}")
        index_pair_label_list = [cols[c] for c in col_indices]
        index, text1, text2, label = index_pair_label_list
        text1 = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in text1.split()])))
        text2 = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in text2.split()])))
        return index, text1, text2, label


@export
@register_reader(task='classify', name='jsonl-paired-shared-vec')
class JSONLIndexPairLabelReader(IndexPairLabelReader):

    def __init__(self, vectorizers, trim=False, truncate=False, **kwargs):
        super().__init__(vectorizers, trim, truncate, **kwargs)
        self.has_header = False

    def index_pair_label(self, line, clean_fn, header):
        cols = json.loads(line)
        index_pair_label_list = [cols[c] for c in self.col_keys]
        index, text1, text2, label = index_pair_label_list
        text1 = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in text1.split()])))
        text2 = ' '.join(list(filter(lambda s: len(s) != 0, [clean_fn(w) for w in text2.split()])))
        return index, text1, text2, label


@export
@register_reader(task='lm', name='default')
class LineSeqReader:

    def __init__(self, vectorizers, trim=False, **kwargs):
        self.nctx = kwargs['nctx']
        self.vectorizers = vectorizers

    def build_vocab(self, files, **kwargs):
        if _all_predefined_vocabs(self.vectorizers):
            logger.info("Skipping building vocabulary.  All vectorizers have predefined vocabs!")
            return {k: v.vocab for k, v in self.vectorizers.items()}

        vocab_file = kwargs.get('vocab_file')
        if vocab_file is not None:
            _vocab_allowed(self.vectorizers)
            return _build_vocab_for_col(0, listify(vocab_file), self.vectorizers)

        vocabs = {k: Counter() for k in self.vectorizers.keys()}

        for file in files:
            if file is None:
                continue

            with codecs.open(file, encoding='utf-8', mode='r') as f:
                sentences = []
                for line in f:
                    sentences += line.split() + ['<EOS>']
                for k, vectorizer in self.vectorizers.items():
                    vocabs[k].update(vectorizer.count(sentences))

        vocabs = _filter_vocab(vocabs, kwargs.get('min_f', {}))
        return vocabs

    def load(self, filename, vocabs, batchsz, tgt_key='x'):

        x = dict()
        with codecs.open(filename, encoding='utf-8', mode='r') as f:
            sentences = []
            for line in f:
                sentences += line.split() + ['<EOS>']
            for k, vectorizer in self.vectorizers.items():
                vec, valid_lengths = vectorizer.run(sentences, vocabs[k])
                x[k] = vec[:valid_lengths]
                shp = list(vectorizer.get_dims())
                shp[0] = valid_lengths
                x['{}_dims'.format(k)] = tuple(shp)

        return baseline.data.SeqWordCharDataFeed(x, self.nctx, batchsz, tgt_key=tgt_key)



import os
import copy
import logging
from itertools import chain
import tensorflow as tf

from baseline.tf.embeddings import *
from eight_mile.tf.layers import *
from baseline.version import __version__

from baseline.utils import (
    fill_y,
    listify,
    ls_props,
    read_json,
    write_json,
    MAGIC_VARS,
    MEAD_HUB_MODULES
)
from baseline.model import DependencyParserModel, register_model
from baseline.tf.tfy import (
    TRAIN_FLAG,
    reload_embeddings,
    new_placeholder_dict,
    tf_device_wrapper,
    create_session,
    BaseLayer,
    TensorDef
)


logger = logging.getLogger('baseline')


class DependencyParserModelBase(tf.keras.Model, DependencyParserModel):
    """Base for all baseline implementations of token-based classifiers

    This class provides a loose skeleton around which the baseline models
    are built.  It is built on the Keras Model base, and fulfills the `DependencyParserModel` interface.
    To override this class, the use would typically override the `create_layers` function which will
    create and attach all sub-layers of this model into the class, and the `call` function which will
    give the model some implementation to call on forward.
    """
    def __init__(self, name=None):
        """Base
        """
        super().__init__(name=name)
        self._unserializable = []

    def set_saver(self, saver):
        self.saver = saver

    def save_values(self, basename):
        """Save tensor files out

        :param basename: Base name of model
        :return:
        """
        self.save_weights(f"{basename}.wgt")

    def save_md(self, basename):
        """This method saves out a `.state` file containing meta-data from these classes and any info
        registered by a user-defined derived class as a `property`. Also write the `graph` and `saver` and `labels`

        :param basename:
        :return:
        """
        write_json(self._state, '{}.state'.format(basename))
        write_json(self.labels, '{}.labels'.format(basename))
        for key, embedding in self.embeddings.items():
            if hasattr(embedding, 'save_md'):
                embedding.save_md('{}-{}-md.json'.format(basename, key))

    def _record_state(self, embeddings: Dict[str, BaseLayer], **kwargs):
        embeddings_info = {}
        for k, v in embeddings.items():
            embeddings_info[k] = v.__class__.__name__

        blacklist = set(chain(
            self._unserializable,
            MAGIC_VARS,
            embeddings.keys(),
            (f'{k}_lengths' for k in embeddings.keys())
        ))
        self._state = {k: v for k, v in kwargs.items() if k not in blacklist}
        self._state.update({
            'version': __version__,
            'module': self.__class__.__module__,
            'class': self.__class__.__name__,
            'embeddings': embeddings_info,
            'hub_modules': MEAD_HUB_MODULES
        })

    def save(self, basename, **kwargs):
        """Save meta-data and actual data for a model

        :param basename: (``str``) The model basename
        :param kwargs:
        :return: None
        """
        self.save_md(basename)
        self.save_values(basename)

    def predict_batch(self, batch_dict):

        batch_dict = self.make_input(batch_dict)
        arcs, rels = self(batch_dict)
        arcs = tf.nn.softmax(arcs).numpy()
        rels = tf.nn.softmax(rels).numpy()
        return arcs, rels

    def predict(self, batch_dict):
        probs = self.predict_batch(batch_dict)
        return probs

    def drop_inputs(self, key, x, do_dropout):
        """Do dropout on inputs, using the dropout value (or none if not set)
        This works by applying a dropout mask with the probability given by a
        value within the `dropin_value: Dict[str, float]`, keyed off the text name
        of the feature

        :param key: The feature name
        :param x: The tensor to drop inputs for
        :param do_dropout: A `bool` specifying if dropout is turned on
        :return: The dropped out tensor
        """
        v = self.dropin_value.get(key, 0)
        if do_dropout and v > 0.0:
            drop_indices = np.where((np.random.random(x.shape) < v) & (x != Offsets.PAD))
            x[drop_indices[0], drop_indices[1]] = Offsets.UNK
        return x

    def make_input(self, batch_dict, train=False):
        """Transform a `batch_dict` into a TensorFlow `feed_dict`

        :param batch_dict: (``dict``) A dictionary containing all inputs to the embeddings for this model
        :param train: (``bool``) Are we training.  Defaults to False
        :return:
        """
        SET_TRAIN_FLAG(train)
        batch_for_model = {}
        for k in self.embeddings.keys():
            batch_for_model[k] = self.drop_inputs(k, batch_dict[k], train)
        # Allow us to track a length, which is needed for BLSTMs
        if self.lengths_key is not None:
            batch_for_model["lengths"] = batch_dict[self.lengths_key]

        return batch_for_model

    def get_labels(self) -> List[str]:
        """Get the string labels back

        :return: labels
        """
        return self.labels

    @classmethod
    @tf_device_wrapper
    def load(cls, basename: str, **kwargs) -> 'DependencyParserModelBase':
        """Reload the model from a graph file and a checkpoint

        The model that is loaded is independent of the pooling and stacking layers, making this class reusable
        by sub-classes.

        :param basename: The base directory to load from
        :param kwargs: See below

        :Keyword Arguments:
        * *sess* -- An optional tensorflow session.  If not passed, a new session is
            created

        :return: A restored model
        """
        _state = read_json("{}.state".format(basename))
        if __version__ != _state['version']:
            logger.warning("Loaded model is from baseline version %s, running version is %s", _state['version'], __version__)


        embeddings_info = _state.pop('embeddings')
        embeddings = reload_embeddings(embeddings_info, basename)
        # If there is a kwarg that is the same name as an embedding object that
        # is taken to be the input of that layer. This allows for passing in
        # subgraphs like from a tf.split (for data parallel) or preprocessing
        # graphs that convert text to indices
        for k in embeddings_info:
            if k in kwargs:
                _state[k] = kwargs[k]
            # TODO: convert labels into just another vocab and pass number of labels to models.
        labels = read_json("{}.labels".format(basename))
        model = cls.create(embeddings, labels, **_state)
        model._state = _state
        model.load_weights(f"{basename}.wgt")
        return model

    @property
    def lengths_key(self) -> str:
        return self._lengths_key

    @lengths_key.setter
    def lengths_key(self, value: str):
        self._lengths_key = value

    @classmethod
    def create(cls, embeddings: Dict[str, BaseLayer], labels, **kwargs) -> 'DependencyParserModelBase':
        """The main method for creating all :class:`DependencyParserBaseModel` types.

        This method typically instantiates a model with pooling and optional stacking layers.
        Many of the arguments provided are reused by each implementation, but some sub-classes need more
        information in order to properly initialize.  For this reason, the full list of keyword args are passed
        to the :method:`pool` and :method:`stacked` methods.

        :param embeddings: This is a dictionary of embeddings, mapped to their numerical indices in the lookup table
        :param labels: This is a list of the `str` labels
        :param kwargs: There are sub-graph specific Keyword Args allowed for e.g. embeddings. See below for known args:

        :Keyword Arguments:
        * *gpus* -- (``int``) How many GPUs to split training across.  If called this function delegates to
            another class `ClassifyParallelModel` which creates a parent graph and splits its inputs across each
            sub-model, by calling back into this exact method (w/o this argument), once per GPU
        * *model_type* -- The string name for the model (defaults to `default`)
        * *sess* -- An optional tensorflow session.  If not passed, a new session is
            created
        * *lengths_key* -- (``str``) Specifies which `batch_dict` property should be used to determine the temporal length
            if this is not set, it defaults to either `word`, or `x` if `word` is also not a feature
        * *finetune* -- Are we doing fine-tuning of word embeddings (defaults to `True`)
        * *mxlen* -- The maximum signal (`x` tensor temporal) length (defaults to `100`)
        * *dropout* -- This indicates how much dropout should be applied to the model when training.
        * *filtsz* -- This is actually a top-level param due to an unfortunate coupling between the pooling layer
            and the input, which, for convolution, requires input padding.

        :return: A fully-initialized tensorflow dependency parser
        """
        model = cls(name=kwargs.get('name'))
        model.lengths_key = kwargs.get('lengths_key')
        model._record_state(embeddings, **kwargs)
        model.pdrop_value = kwargs.get('dropout', 0.5)
        model.labels = labels["labels"]
        model.punct = labels["labels"].get("punct", Offsets.PAD)
        model.create_layers(embeddings, **kwargs)
        return model

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        """This method defines the model itself, and must be overloaded by derived classes

        This function will update `self` with the layers required to execute the `call()` method

        :param embeddings: The input feature indices
        :param kwargs:
        :return:
        """

    def decode(self, example, **kwargs):
        heads_pred, labels_pred = self(example)
        heads_pred = heads_pred.numpy()
        labels_pred = labels_pred.numpy()
        # Just do a quick greedy decode, pick the argmax of the heads, and for that head, pick the
        # argmax of the label
        B = labels_pred.shape[0]
        T = labels_pred.shape[1]

        # If there is padding, rip it off to the max sequence length so the tensors are the same size
        greedy_heads_pred = np.argmax(heads_pred, -1).reshape(-1)
        greedy_labels_pred = labels_pred.reshape(B * T, T, -1)[
            np.arange(len(greedy_heads_pred)), greedy_heads_pred].reshape(B, T, -1)

        greedy_labels_pred = np.argmax(greedy_labels_pred, -1)
        greedy_heads_pred = greedy_heads_pred.reshape(B, T)
        return greedy_heads_pred, greedy_labels_pred


@register_model(task='deps', name='default')
class BiAffineDependencyParser(DependencyParserModelBase):

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        self.embeddings = self.init_embed(embeddings, **kwargs)
        self.pool_model = self.init_pool(self.embeddings.output_dim, **kwargs)
        output_dim_arcs = kwargs.get('hsz_arcs', 500)
        self.arc_h = self.init_proj(output_dim_arcs, **kwargs)
        self.arc_d = self.init_proj(output_dim_arcs, **kwargs)
        output_dim_rels = kwargs.get('hsz_rels', 100)
        self.rel_h = self.init_proj(output_dim_rels, **kwargs)
        self.rel_d = self.init_proj(output_dim_rels, **kwargs)
        self.arc_attn = self.init_biaffine(self.arc_h.output_dim, 1, True, False)
        self.rel_attn = self.init_biaffine(self.rel_h.output_dim, len(self.labels), True, True)
        self.primary_key = self.lengths_key.split('_')[0]


    def init_proj(self, output_dim: int, **kwargs) -> BaseLayer:
        """Produce a stacking operation that will be used in the model

        :param input_dim: The input dimension size
        :param kwargs:
        :return: A stacking operation (or None)
        """
        return WithDropout(tf.keras.layers.Dense(output_dim, activation=get_activation(kwargs.get('activation', 'leaky_relu'))), pdrop=self.pdrop_value)

    def init_biaffine(self, input_dim: int, output_dim: int, bias_x: bool, bias_y: bool):
        return BilinearAttention(input_dim, output_dim, bias_x, bias_y)

    def init_embed(self, embeddings: Dict[str, TensorDef], **kwargs) -> BaseLayer:
        """This method creates the "embedding" layer of the inputs, with an optional reduction

        :param embeddings: A dictionary of embeddings
        :param kwargs: See below

        :Keyword Arguments:
        * *embeddings_reduction* (defaults to `concat`) An operator to perform on a stack of embeddings
        * *embeddings_name* (``str``) Optional override to Keras default names
        :return: The output of the embedding stack followed by its reduction.  This will typically be an output
          with an additional dimension which is the hidden representation of the input
        """
        reduction = kwargs.get('embeddings_reduction', 'concat')
        name = kwargs.get('embeddings_name')
        return EmbeddingsStack(embeddings, self.pdrop_value, reduction=reduction, name=name)

    def init_pool(self, input_dim: int, **kwargs) -> BaseLayer:
        """Produce a pooling operation that will be used in the model

        :param input_dim: The input dimension size
        :param kwargs:
        :return: A pooling operation
        """
        hsz = kwargs.get('rnnsz', kwargs.get('hsz', 800))
        layers = kwargs.get('layers', 3)
        return BiLSTMEncoderSequence(input_dim, hsz, layers, self.pdrop_value) #, variational=True)

    def init_stacked(self, input_dim: int, **kwargs) -> BaseLayer:
        """Produce a stacking operation that will be used in the model

        :param input_dim: The input dimension size
        :param kwargs: See below

        :keyword arguments:

        * *hsz* (``list``), defaults to nothing, in which case this function is pass-through
        * *stacked_name* (``str``) Optional override to stacking name

        :return: A stacking operation (or None)
        """
        hszs = listify(kwargs.get('hsz', []))
        if not hszs:
            return PassThru(input_dim)
        name = kwargs.get('stacked_name')
        return DenseStack(input_dim, hszs, pdrop_value=self.pdrop_value, name=name)

    def init_output(self, **kwargs):
        """Provide a projection from the encoder output to the number of labels

        This projection typically will not include any activation, since its output is the logits that
        the decoder is built on

        :param kwargs: See below

        :keyword arguments:
        * *output_name* (``str``) Optional override to default Keras layer name
        :return: A projection from the encoder output size to the final number of labels
        """
        name = kwargs.get('output_name')
        return tf.keras.layers.Dense(len(self.labels), name=name)

    def call(self, inputs: Dict[str, TensorDef]) -> Tuple[TensorDef, TensorDef]:
        """Forward execution of the model.  Sub-classes typically shouldnt need to override

        :param inputs: An input dictionary containing the features and the primary key length
        :return: A tensor
        """

        lengths = inputs.get("lengths")
        mask = tf.sequence_mask(lengths)
        embedded = self.embeddings(inputs)
        embedded = (embedded, lengths)
        pooled = self.pool_model(embedded)
        arcs_h = self.arc_h(pooled)
        arcs_d = self.arc_d(pooled)
        rels_h = self.rel_h(pooled)
        rels_d = self.rel_d(pooled)
        score_arcs = self.arc_attn(arcs_d, arcs_h, mask)
        mask = tf.expand_dims(mask, 1)
        score_rels = tf.transpose(self.rel_attn(rels_d, rels_h, mask), [0, 2, 3, 1])
        return score_arcs, score_rels




import time
import logging
import torch
import numpy as np
from baseline.progress import create_progress_bar
from eight_mile.utils import listify
from baseline.utils import get_model_file, get_metric_cmp, convert_seq2seq_golds, convert_seq2seq_preds
from baseline.train import Trainer, create_trainer, register_trainer, register_training_func
from eight_mile.pytorch.optz import OptimizerManager
from eight_mile.bleu import bleu
from baseline.model import create_model_for
from torch.utils.data import DataLoader

logger = logging.getLogger('baseline')


@register_trainer(task='seq2seq', name='default')
class Seq2SeqTrainerPyTorch(Trainer):

    def __init__(self, model, **kwargs):
        super().__init__()
        if type(model) is dict:
            model = create_model_for('seq2seq', checkpoint=kwargs.get('checkpoint'), **model)

        self.clip = float(kwargs.get('clip', 5))
        self.model = model
        self.optimizer = OptimizerManager(self.model, **kwargs)
        self._input = model.make_input
        self._predict = model.predict
        self.tgt_rlut = kwargs['tgt_rlut']
        self.gpus = kwargs.get('gpus', 1)
        self.bleu_n_grams = int(kwargs.get("bleu_n_grams", 4))
        self.label_smoothing = kwargs.get("label_smoothing")

        if self.gpus > 0:
            self.crit = model.create_loss(label_smooth=self.label_smoothing).cuda()
            if self.gpus > 1:
                self.model = torch.nn.DataParallel(model).cuda()
            else:
                self.model.cuda()
        else:
            logger.warning("Requested training on CPU.  This will be slow.")
            self.crit = model.create_loss()

        self.nsteps = kwargs.get('nsteps', 500)

    @staticmethod
    def _num_toks(tgt_lens):
        return torch.sum(tgt_lens).item()

    @staticmethod
    def _acc(preds, golds):
        """Calculate the accuracy of exact matching between preds and golds. This metric is particularly useful when
        using Seq2SeqModel for prediction."""
        total = len(preds)
        correct = 0
        for pred, gold in zip(preds, golds):
            if pred == gold[0]:
                correct += 1
        return float(correct)/total

    def save(self, model_file):
        self._get_pytorch_model().save(model_file)

    def _get_pytorch_model(self):
        return self.model.module if self.gpus > 1 else self.model

    def calc_metrics(self, agg, norm):
        metrics = super().calc_metrics(agg, norm)
        metrics['perplexity'] = np.exp(metrics['avg_loss'])
        return metrics

    def test(self, vs, reporting_fns, phase, **kwargs):
        if phase == 'Test':
            return self._evaluate(vs, reporting_fns, **kwargs)

        self.model.eval()
        total_loss = total_toks = 0
        steps = len(vs)
        self.valid_epochs += 1
        preds = []
        golds = []

        start = time.time()
        pg = create_progress_bar(steps)
        for batch_dict in pg(vs):
            input_ = self._input(batch_dict)
            tgt = input_['tgt']
            tgt_lens = input_['tgt_len']
            pred = self.model(input_)
            loss = self.crit(pred, tgt)
            toks = self._num_toks(tgt_lens)
            total_loss += loss.item() * toks
            total_toks += toks
            greedy_preds = [p[0] for p in self._predict(input_, beam=1, make_input=False)[0]]
            preds.extend(convert_seq2seq_preds(greedy_preds, self.tgt_rlut))
            golds.extend(convert_seq2seq_golds(tgt.cpu().numpy(), tgt_lens, self.tgt_rlut))

        metrics = self.calc_metrics(total_loss, total_toks)
        metrics['bleu'] = bleu(preds, golds, self.bleu_n_grams)[0]
        metrics['acc'] = self._acc(preds, golds)
        self.report(
            self.valid_epochs, metrics, start,
            phase, 'EPOCH', reporting_fns
        )
        return metrics

    def _evaluate(self, es, reporting_fns, **kwargs):
        self.model.eval()
        pg = create_progress_bar(len(es))
        preds = []
        golds = []
        start = time.time()
        for batch_dict in pg(es):
            tgt = batch_dict['tgt']
            tgt_lens = batch_dict['tgt_lengths']
            pred = [p[0] for p in self._predict(batch_dict, numpy_to_tensor=False, **kwargs)[0]]
            preds.extend(convert_seq2seq_preds(pred, self.tgt_rlut))
            golds.extend(convert_seq2seq_golds(tgt, tgt_lens, self.tgt_rlut))
        metrics = {'bleu': bleu(preds, golds, self.bleu_n_grams)[0]}
        metrics['acc'] = self._acc(preds, golds)
        self.report(
            0, metrics, start, 'Test', 'EPOCH', reporting_fns
        )
        return metrics

    def train(self, ts, reporting_fns):
        self.model.train()

        epoch_loss = 0
        epoch_toks = 0

        start = time.time()
        self.nstep_start = start
        for batch_dict in ts:

            start_time = time.time()
            self.optimizer.zero_grad()
            input_ = self._input(batch_dict)
            tgt = input_['tgt']
            pred = self.model(input_)
            loss = self.crit(pred, tgt)
            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.clip)
            self.optimizer.step()
            tgt_lens = batch_dict['tgt_lengths']
            tok_count = self._num_toks(tgt_lens)
            reporting_loss = loss.item() * tok_count
            epoch_loss += reporting_loss
            epoch_toks += tok_count
            self.nstep_agg += reporting_loss
            self.nstep_div += tok_count

            if (self.optimizer.global_step + 1) % self.nsteps == 0:
                metrics = self.calc_metrics(self.nstep_agg, self.nstep_div)
                metrics['lr'] = self.optimizer.current_lr
                self.report(
                    self.optimizer.global_step + 1, metrics, self.nstep_start,
                    'Train', 'STEP', reporting_fns, self.nsteps
                )
                self.reset_nstep()

        metrics = self.calc_metrics(epoch_loss, epoch_toks)
        metrics['lr'] = self.optimizer.current_lr

        self.train_epochs += 1
        self.report(
            self.train_epochs, metrics, start,
            'Train', 'EPOCH', reporting_fns
        )
        return metrics


@register_training_func('seq2seq')
def fit(model_params, ts, vs, es=None, **kwargs):

    do_early_stopping = bool(kwargs.get('do_early_stopping', True))
    epochs = int(kwargs.get('epochs', 20))
    model_file = get_model_file('seq2seq', 'pytorch', kwargs.get('basedir'))


    num_loader_workers = int(kwargs.get('num_loader_workers', 0))
    pin_memory = bool(kwargs.get('pin_memory', True))

    if not isinstance(ts, DataLoader):
        ts = DataLoader(ts, num_workers=num_loader_workers, batch_size=None, pin_memory=pin_memory)
    if not isinstance(vs, DataLoader):
        vs = DataLoader(vs, batch_size=None, pin_memory=pin_memory)
    if es and not isinstance(es, DataLoader):
        es = DataLoader(es, batch_size=None, pin_memory=pin_memory)

    best_metric = 0
    if do_early_stopping:
        early_stopping_metric = kwargs.get('early_stopping_metric', 'perplexity')
        early_stopping_cmp, best_metric = get_metric_cmp(early_stopping_metric, kwargs.get('early_stopping_cmp'))
        patience = kwargs.get('patience', epochs)
        logger.info('Doing early stopping on [%s] with patience [%d]', early_stopping_metric, patience)

    reporting_fns = listify(kwargs.get('reporting', []))
    logger.info('reporting %s', reporting_fns)

    after_train_fn = kwargs.get('after_train_fn', None)
    trainer = create_trainer(model_params, **kwargs)

    last_improved = 0
    for epoch in range(epochs):
        trainer.train(ts, reporting_fns)

        if after_train_fn is not None:
            after_train_fn(trainer.model)

        test_metrics = trainer.test(vs, reporting_fns, phase='Valid')

        if do_early_stopping is False:
            trainer.save(model_file)

        elif early_stopping_cmp(test_metrics[early_stopping_metric], best_metric):
            last_improved = epoch
            best_metric = test_metrics[early_stopping_metric]
            logger.info('New best %.3f', best_metric)
            trainer.save(model_file)

        elif (epoch - last_improved) > patience:
            logger.info('Stopping due to persistent failures to improve')
            break

    if do_early_stopping is True:
        logger.info('Best performance on %s: %.3f at epoch %d', early_stopping_metric, best_metric, last_improved)

    if es is not None:
        model = torch.load(model_file)
        trainer = Seq2SeqTrainerPyTorch(model, **kwargs)
        test_metrics = trainer.test(es, reporting_fns, phase='Test')
    return test_metrics



import logging
from baseline.model import DependencyParserModel, register_model
from baseline.pytorch.torchy import *
from eight_mile.pytorch.layers import sequence_mask_mxlen, truncate_mask_over_time
from baseline.utils import listify, write_json, revlut
from eight_mile.pytorch.layers import *
import torch.backends.cudnn as cudnn
import torch.jit as jit
import os
cudnn.benchmark = True

logger = logging.getLogger('baseline')


def decode_results(heads_pred: TensorDef, labels_pred: TensorDef) -> Tuple[TensorDef, TensorDef]:
    # Just do a quick greedy decode, pick the argmax of the heads, and for that head, pick the
    # argmax of the label
    B = labels_pred.shape[0]
    T = labels_pred.shape[1]

    # If there is padding, rip it off to the max sequence length so the tensors are the same size
    greedy_heads_pred = torch.argmax(heads_pred, -1).view(-1)
    greedy_labels_pred = labels_pred.reshape(B * T, T, -1)[
        torch.arange(len(greedy_heads_pred)), greedy_heads_pred].view(B, T, -1)

    greedy_labels_pred = torch.argmax(greedy_labels_pred, -1)
    greedy_heads_pred = greedy_heads_pred.view(B, T)
    return greedy_heads_pred, greedy_labels_pred


class ArcLabelLoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.arc_loss = nn.CrossEntropyLoss(ignore_index=-1)
        self.label_loss = nn.CrossEntropyLoss(ignore_index=0)

    def forward(self, arcs_pred, arcs_gold, labels_pred, labels_gold):
        # First, trim gold labels to length of the input
        B = labels_pred.shape[0]
        T = labels_pred.shape[1]
        flat_labels_pred = labels_pred.reshape(B*T, T, -1)
        flat_arcs_pred = arcs_pred.reshape(B*T, T)
        flat_labels_gold = labels_gold[:, :T].contiguous().view(-1)
        flat_arcs_gold = arcs_gold[:, :T].contiguous().view(-1)
        flat_labels_pred = flat_labels_pred[torch.arange(len(flat_arcs_gold)), flat_arcs_gold]
        arc_loss = self.arc_loss(flat_arcs_pred, flat_arcs_gold)
        rel_loss = self.label_loss(flat_labels_pred, flat_labels_gold)
        return arc_loss + rel_loss


class DependencyParserModelBase(nn.Module, DependencyParserModel):

    def __init__(self):
        super().__init__()
        self.gpu = False

    @classmethod
    def load(cls, filename: str, **kwargs) -> 'DependencyParserModelBase':
        device = kwargs.get('device')
        if not os.path.exists(filename):
            filename += '.pyt'
        model = torch.load(filename, map_location=device)
        model.gpu = False if device == 'cpu' else model.gpu
        return model

    def save(self, outname: str):
        logger.info('saving %s' % outname)
        torch.save(self, outname)
        basename, _ = os.path.splitext(outname)
        write_json(self.labels, basename + ".labels")

    @classmethod
    def create(cls, embeddings, labels, **kwargs) -> 'DependencyParserModelBase':

        model = cls()
        model.pdrop = kwargs.get('pdrop', 0.333)
        model.dropin_values = kwargs.get('dropin', {})
        model.lengths_key = kwargs.get('lengths_key')
        model.gpu = not bool(kwargs.get('nogpu', False))
        model.labels = labels["labels"]
        model.punct = labels["labels"].get("punct", Offsets.PAD)
        model.create_layers(embeddings, **kwargs)
        logger.info(model)
        return model

    def cuda(self, device=None):
        self.gpu = True
        return super().cuda(device=device)

    def create_loss(self):
        return ArcLabelLoss()

    def drop_inputs(self, key, x):
        """Do dropout on inputs, using the dropout value (or none if not set)
        This works by applying a dropout mask with the probability given by a
        value within the `dropin_values: Dict[str, float]`, keyed off the text name
        of the feature

        :param key: The feature name
        :param x: The tensor to drop inputs for
        :return: The dropped out tensor
        """
        v = self.dropin_values.get(key, 0)
        if not self.training or v == 0:
            return x

        mask_pad = x != Offsets.PAD
        mask_drop = x.new(x.size(0), x.size(1)).bernoulli_(v).to(mask_pad.dtype)
        x.masked_fill_(mask_pad & mask_drop, Offsets.UNK)
        return x

    def input_tensor(self, key, batch_dict, perm_idx, numpy_to_tensor=False):
        """Given a batch of input, and a key, prepare and noise the input

        :param key: The key of the tensor
        :param batch_dict: The batch of data as a dictionary of tensors
        :param perm_idx: The proper permutation order to get lengths descending
        :param numpy_to_tensor: Should this tensor be converted to a `torch.Tensor`
        :return:
        """
        tensor = batch_dict[key]
        if numpy_to_tensor:
            tensor = torch.from_numpy(tensor)

        tensor = self.drop_inputs(key, tensor)
        tensor = tensor[perm_idx]
        if self.gpu:
            tensor = tensor.cuda()
        return tensor

    def make_input(self, batch_dict, perm=False, numpy_to_tensor=False):

        """Transform a `batch_dict` into something usable in this model

        :param batch_dict: (``dict``) A dictionary containing all inputs to the embeddings for this model
        :return:
        """
        example_dict = dict({})
        perm_idx = None

        # Allow us to track a length, which is needed for BLSTMs
        if self.lengths_key is not None:
            lengths = batch_dict[self.lengths_key]
            if numpy_to_tensor:
                lengths = torch.from_numpy(lengths)
            lengths, perm_idx = lengths.sort(0, descending=True)
            if self.gpu:
                lengths = lengths.cuda()
            example_dict['lengths'] = lengths

        for key in self.embeddings.keys():
            example_dict[key] = self.input_tensor(key, batch_dict, perm_idx, numpy_to_tensor=numpy_to_tensor)

        y = batch_dict.get('heads')
        if y is not None:
            if numpy_to_tensor:
                y = torch.from_numpy(y)

            if perm_idx is not None:
                y = y[perm_idx]
            if self.gpu:
                y = y.cuda()
            example_dict['heads'] = y

        y = batch_dict.get('labels')
        if y is not None:
            if numpy_to_tensor:
                y = torch.from_numpy(y)
            if perm_idx is not None:
                y = y[perm_idx]
            if self.gpu:
                y = y.cuda()
            example_dict['labels'] = y

        if perm:
            return example_dict, perm_idx

        return example_dict

    def predict_batch(self, batch_dict: Dict[str, TensorDef], **kwargs) -> TensorDef:
        numpy_to_tensor = bool(kwargs.get('numpy_to_tensor', True))
        decode = bool(kwargs.get('decode', True))
        examples, perm_idx = self.make_input(batch_dict, perm=True, numpy_to_tensor=numpy_to_tensor)
        with torch.no_grad():
            if decode:
                arcs, rels = self.decode(examples)

            else:
                arcs, rels = self(examples)
                arcs.exp_()
                rels.exp_()
            arcs = unsort_batch(arcs, perm_idx)
            rels = unsort_batch(rels, perm_idx)

        return arcs, rels

    def predict(self, batch_dict: Dict[str, TensorDef], **kwargs):
        """Raw prediction, for non-fancy decoding, use `def decode()` instead

        :param batch_dict:
        :param kwargs:
        :return:
        """
        self.eval()
        arcs, rels = self.predict_batch(batch_dict, **kwargs)
        return arcs, rels

    def get_labels(self) -> List[str]:
        return self.labels

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        """This method defines the model itself, and must be overloaded by derived classes

        This function will update `self` with the layers required to execute the `call()` method

        :param embeddings: The input feature indices
        :param kwargs:
        :return:
        """

    def decode(self, example, **kwargs):
        heads_pred, labels_pred = self(example)
        return decode_results(heads_pred, labels_pred)


@register_model(task='deps', name='default')
class BiAffineDependencyParser(DependencyParserModelBase):

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        self.embeddings = self.init_embed(embeddings, **kwargs)
        self.pool_model = self.init_pool(self.embeddings.output_dim, **kwargs)
        output_dim_arcs = kwargs.get('hsz_arcs', 500)
        self.arc_h = self.init_proj(self.pool_model.output_dim, output_dim_arcs, **kwargs)
        self.arc_d = self.init_proj(self.pool_model.output_dim, output_dim_arcs, **kwargs)
        output_dim_rels = kwargs.get('hsz_rels', 100)
        self.rel_h = self.init_proj(self.pool_model.output_dim, output_dim_rels, **kwargs)
        self.rel_d = self.init_proj(self.pool_model.output_dim, output_dim_rels, **kwargs)
        self.arc_attn = self.init_biaffine(self.arc_h.output_dim, 1, True, False)
        self.rel_attn = self.init_biaffine(self.rel_h.output_dim, len(self.labels), True, True)
        self.primary_key = self.lengths_key.split('_')[0]

    def init_embed(self, embeddings: Dict[str, TensorDef], **kwargs) -> BaseLayer:
        """This method creates the "embedding" layer of the inputs, with an optional reduction

        :param embeddings: A dictionary of embeddings

        :Keyword Arguments: See below
        * *embeddings_reduction* (defaults to `concat`) An operator to perform on a stack of embeddings
        * *embeddings_dropout = float(kwargs.get('embeddings_dropout', 0.0))

        :return: The output of the embedding stack followed by its reduction.  This will typically be an output
          with an additional dimension which is the hidden representation of the input
        """
        reduction = kwargs.get('embeddings_reduction', 'concat')
        embeddings_dropout = float(kwargs.get('embeddings_dropout', self.pdrop))
        return EmbeddingsStack(embeddings, embeddings_dropout, reduction=reduction)

    def init_pool(self, input_dim: int, **kwargs) -> BaseLayer:
        """Produce a pooling operation that will be used in the model

        :param input_dim: The input dimension size
        :param kwargs:
        :return: A pooling operation
        """
        hsz = kwargs.get('rnnsz', kwargs.get('hsz', 800))
        layers = kwargs.get('layers', 3)
        if layers == 0:
            logger.warning("No layers given. Setting up pass-through model")
            return WithoutLength(PassThru(input_dim))
        return BiLSTMEncoderSequence(input_dim, hsz, layers, self.pdrop, batch_first=True, initializer="ortho")


    def init_proj(self, input_dim: int, output_dim: int, **kwargs) -> BaseLayer:
        """Produce a stacking operation that will be used in the model

        :param input_dim: The input dimension size
        :param kwargs:
        :return: A stacking operation (or None)
        """
        return WithDropout(Dense(input_dim, output_dim, activation=kwargs.get('activation', 'leaky_relu'), initializer="ortho"),
                           pdrop=self.pdrop)

    def init_biaffine(self, input_dim: int, output_dim: int, bias_x: bool, bias_y: bool):
        return BilinearAttention(input_dim, output_dim, bias_x, bias_y)



    def forward(self, inputs: Dict[str, TensorDef]) -> TensorDef:
        """Forward execution of the model.  Sub-classes typically shouldnt need to override

        :param inputs: An input dictionary containing the features and the primary key length
        :return: A tensor
        """

        lengths = inputs.get("lengths")
        Tin = inputs[self.primary_key].shape[1]
        mask = sequence_mask_mxlen(lengths, max_len=Tin).to(lengths.device)
        embedded = self.embeddings(inputs)
        embedded = (embedded, lengths)
        pooled = self.pool_model(embedded)
        arcs_h = self.arc_h(pooled)
        arcs_d = self.arc_d(pooled)
        rels_h = self.rel_h(pooled)
        rels_d = self.rel_d(pooled)
        mask = truncate_mask_over_time(mask, arcs_h)
        score_arcs = self.arc_attn(arcs_d, arcs_h, mask)
        score_rels = self.rel_attn(rels_d, rels_h, mask.unsqueeze(1)).permute(0, 2, 3, 1)
        return score_arcs, score_rels
